;; Problem 30
;; 08 November 2002

;; Surprisingly there are only three numbers that can be written as
;; the sum of fourth powers of their digits:

;;     1634 = 1^4 + 6^4 + 3^4 + 4^4
;;     8208 = 8^4 + 2^4 + 0^4 + 8^4
;;     9474 = 9^4 + 4^4 + 7^4 + 4^4

;; As 1 = 1^4 is not a sum it is not included.

;; The sum of these numbers is 1634 + 8208 + 9474 = 19316.

;; Find the sum of all the numbers that can be written as the sum of fifth powers of their digits.

(defun char->digit (chr)
  (cond
    ((eq #\1 chr) 1)
    ((eq #\2 chr) 2)
    ((eq #\3 chr) 3)
    ((eq #\4 chr) 4)
    ((eq #\5 chr) 5)
    ((eq #\6 chr) 6)
    ((eq #\7 chr) 7)
    ((eq #\8 chr) 8)
    ((eq #\9 chr) 9)
    ((eq #\0 chr) 0)
    (t
     (error "unsupported char: ~a~&" chr))))

; (char->digit (char "1" 0))

(defun explode-num (num)
  (let ((str (format nil "~d" num)))
    (loop for ii from 0 to (- (length str) 1)
	  collect (char->digit (char str ii)))))

;; (explode-num 1634) ; => (1 6 3 4)

(defun sum-of-powers (num pow)
  (declare (optimize speed)
	   (type fixnum num)
	   (type fixnum pow))
  (let* ((parts (explode-num num))
	 (nths (mapcar #'(lambda (n) (expt n pow)) parts)))
;;     (format t "pow:~a num:~a =~a ~a => ~a~&"
;; 	    pow
;; 	    num
;; 	    (apply #'+ nths)
;; 	    parts
;; 	    nths)
    (apply #'+ nths)))

(defun sum-of-powers2 (num pow)
  (declare (optimize speed)
	   (type fixnum num)
	   (type fixnum pow))
  (let ((sum 0))
    (declare (type fixnum sum))
    (loop for nn in (explode-num num)
	  do
	  (setf sum (+ sum (expt nn pow))))
    sum))

;(time (loop for ii from 2 to 200000 do (sum-of-powers ii 5)))
;(time (loop for ii from 2 to 200000 do (sum-of-powers2 ii 5)))

;;(sum-of-powers 1634 4)   pow:4 num:1634 =1634 (1 6 3 4) => (1 1296 81 256)
;;(sum-of-powers 10634 4)  pow:4 num:10634 =1634 (1 0 6 3 4) => (1 0 1296 81 256)

;; (sum-of-powers 100 4) ;; pow:4 num:100 =1 (1 0 0) => (1 0 0)


'(loop for ii from 1 to 100
      do
      (sum-of-powers ii 3))

'(loop for ii from 2 to 2000
      do
      (when (= ii (sum-of-powers ii 2))
	(format t "FOUND: ~a~&" ii)))




;; eventually the number will outpace the ability of all 9's ^ pow to
;; 'catch up' with it

'(loop for ii in '(9 99 999 9999 99999 999999 9999999 99999999)
      do
      (when (> ii (sum-of-powers ii 5))
	(format t "OVER~&")))

(defun find-fifth-powers (&optional (max 999999))
  (let ((res (list)))
    (loop for ii from 2 to max
	  do
	  (when (= 0 (mod ii 1000))
	    (format t "."))
	  (when (= ii (sum-of-powers2 ii 5))
	    (format t "found:~a~&" ii)
	    (push ii res)))
    (format t "res:~a~&" res)
    res))

(time (find-fifth-powers))

(+ 194979 93084 92727 54748 4151 4150) ; 443839